import numpy as np
import math
import scipy.io.wavfile as wav
from scipy.signal import lfilter
from copy import deepcopy
import matplotlib.pyplot as plt
import seaborn as sns

#Colors
colors = plt.get_cmap('tab20')(np.arange(0,2));cmap = sns.color_palette([colors[1,:], colors[0,:]])


def speech_wave(INPUT_PATH):
    """Read a 16-bit or 32-bit audio (.wav) file. Device each samples by the number of bits used. 
    Average the channels if multiple ones.

    Parameters
    ----------
    INPUT_PATH : str
        The audio file location.
    print_cols : bool, optional
        A flag used to print the columns to the console (default is
        False)

    Returns
    -------
    frames: int
        Sampling frequency.
    signal: np.array dtype('float64')
        Audio signal with possibly multiple channels.
    """
    
    (fs,signal) = wav.read(INPUT_PATH)
    if signal.dtype == 'int16':
        nb = 16 # -> 16-bit wav files
    elif signal.dtype == 'int32':
        nb = 32 # -> 32-bit wav files
    max_nb = float(2 ** (nb - 1))
    signal = signal / (max_nb + 1.0)  
    
    if len(signal.shape) > 1:
        signal = signal[:,0]
    return fs, signal
 
def create_frames(signal, fs, FRAME_LENGTH, FRAME_DISTANCE):
    """Create array of frames of a 1D signal, specyfying length of 
    a frame and delay betwwen frames.

    Parameters
    ----------
    signal : 1D array.
        Array of the audio file.
    fs : int
        Sampling Frequency of the audio signal.
    FRAME_LENGTH: float.
        Length of a frame in seconds.
    FRAME_DISTANCE: float.
        Delay between each frame.

    Returns
    -------
    frames: np.array dtype('int64')
        Array containing each frames extracted from the aduio signal.
        
    """
    
    N = len(signal) # Number of samples
    samples_per_frame = int(np.fix(fs*FRAME_LENGTH)) # Number of samples per frame 
    samples_between_frames = int(np.fix(fs*FRAME_DISTANCE)) # Number of samples between each frame
     
    if len(signal) < samples_per_frame:
        print("speech file length shorter than window length")
        exit()
     
    # Number of frames using ceil 
    frames = int(np.ceil( (N - samples_per_frame + samples_between_frames)/samples_between_frames))
    
    # Required signal length
    slen = (frames-1)*samples_between_frames + samples_per_frame

    # If the signal is too short, it is filled with zeros
    if len(signal) < slen:
        signal = np.concatenate((signal, np.zeros((slen - N))))
    else:
        signal = deepcopy(signal)
  
    # Do the frame extraction based on numpy slicing 
    samples_idx_for_each_frames = np.tile(np.arange(0,samples_per_frame),(frames,1)) + np.tile(np.arange(0,(frames)*samples_between_frames,samples_between_frames),(samples_per_frame,1)).T
    samples_idx_for_each_frames = np.array(samples_idx_for_each_frames, dtype=np.int64)
    return signal[samples_idx_for_each_frames]

def compute_spectral_flatness(signal, fs, FRAME_LENGTH, FRAME_DISTANCE, nftt):
    """Compute the spectral flatness of each frames

    Parameters
    ----------
    signal : 1D array.
        Array of the audio file.
    fs : int
        Sampling Frequency of the audio signal.
    FRAME_LENGTH: float.
        Length of a frame in seconds.
    FRAME_DISTANCE: float.
        Delay between each frame.
    nftt: int
        Length of the output fft signal.
        

    Returns
    -------
    spectral_flatness: 1D array dtype('float32')
        1D array of the spectral flatness for each frames
    samples_per_frame: int
        Samples per frames.
    samples_between_frames: int
        Samples between frames.
    n_frames_fft: int
        Number of frames (floor)

    """
    
    eps=np.finfo(float).eps

    # Create frames from the input signal.
    xf = create_frames(signal, fs, FRAME_LENGTH, FRAME_DISTANCE) 
    
    # Define Hamming window and tile it so that it has the shape of the frames array.
    w = np.matrix(np.hamming(int(fs*FRAME_LENGTH)) )
    w = np.tile(w,(np.size(xf, axis=0), 1))

    # Apply hamming window to the frames
    xf = np.multiply (xf, w) 
    
    # Compute the one-dimensional discrete Fourier Transform of each frames
    ak=np.abs(np.fft.fft(xf,nftt))
    
    # Subsample fft coefficients.
    idx = range(0,int(nftt/2) +1)
    ak=ak[:,idx]
    
    # Normalised exponential of the sum of the log of the coeeficients for each frames. 
    # leads to one number (numerator) for each frames
    Num=np.exp( float(1/len(idx)) * np.sum(np.log(ak+eps), axis=1) ) 
    
    # Normalised  sum of the coeeficients for each frames. 
    # leads to one number (denominator) for each frames
    Den=float(1/len(idx)) * np.sum(ak, axis=1)
    
    # Compute ratio of Num over Den
    spectral_flatness=(Num+eps)/(Den+eps)

    samples_per_frame = int(np.fix(fs*FRAME_LENGTH)) # Number of samples per frame 
    samples_between_frames = int(np.fix(fs*FRAME_DISTANCE)) # Number of samples between each frame
     
    # Number of frames using floor 
    n_frames_fft =int(np.floor((len(signal)-(samples_per_frame-samples_between_frames))/samples_between_frames))

    # Change the size of spectral_flatness so that it is the number of frames
    if n_frames_fft < len(spectral_flatness):
        spectral_flatness=spectral_flatness[range(n_frames_fft)]
    else:
        spectral_flatness = np.concatenate((spectral_flatness, np.repeat(spectral_flatness[:1], n_frames_fft -len(spectral_flatness), axis=0) ))

    return spectral_flatness, samples_per_frame, samples_between_frames, n_frames_fft

def classify_pitch_frames(spectral_flatness, n_frames_fft, SPECTRAL_FLATNESS_THRESHOLD):
    """Classify the frames of the signal depending on whether their spectral flattness is 
        lower than the `SPECTRAL_FLATNESS_THRESHOLD` threshold.
    
    Parameters
    ----------
    spectral_flatness : 1D array. 
        Spectral flatness of the signal, that is considered as the pitch of the audio signal.
    n_frames_fft: int.
        Number of frames of the fft signal (floor).
    SPECTRAL_FLATNESS_THRESHOLD: float default=.5
        Spectral flatness threshold aboce which a frame is labelled as noise. 

    Returns
    -------
    spectral_flatness_bin_mask : 1D binary array.
        Mask of size n_frames_fft. 1 if the frame has lower spectral flatness than the threshold (e.g. 05), 0 otherwise. 
        
    """
    # Init. the mask on the frames. 0 if no speech a priori. 1 if pitch.
    spectral_flatness_bin_mask=np.zeros(n_frames_fft)

    # Put 1 if the spectral flatness is lower than the threshold of .5. (theta_sft in the paper.)
    spectral_flatness_bin_mask[np.less_equal(spectral_flatness, SPECTRAL_FLATNESS_THRESHOLD)]=1 

    return spectral_flatness_bin_mask

def spectral_flatness_block_extension(spectral_flatness_bin_mask, spectral_flatness, n_frames_fft, POST_PROCESS_SPECTRAL_FLATNESS_BIN_MASK):
    """Extend the binary pitch mask by 60 frames. 
    From section 3.3 of the paper:Extended pitch segment detection 
    The VAD algorithm is based on the denoised speech and the pitch 
    information generated from the previous steps. The fundamental assumption is that 
    all speech segments should contain a number of speech frames with pitch. In this algorithm, 
    pitch frames are ﬁrst grouped into pitch segments, which are then extended from both ends by 
    60 frames (600ms), based on speech statistics, in order to include voiced sounds, unvoiced 
    sounds and potentially non-speech parts. This strategy is taken for the following reasons: 
    1) pitch information is already extracted in the previous steps, 2) pitch is used as an anchor 
    to trigger the VAD process and 3) many frames can be potentially discarded if an utterance
    contains a large portion of non-speech segments, which are also non-pitch segments.
        
    
    Parameters
    ----------
    spectral_flatness_bin_mask : 1D binary array.
        Mask of size n_frames_fft. 1 if the frame has lower spectral flatness than the threshold (e.g. 05), 0 otherwise. 
    spectral_flatness : 1D array. 
        Spectral flatness of the signal, that is considered as the pitch of the audio signal.
    n_frames_fft: int.
        Number of frames of the fft signal (floor).
    POST_PROCESS_SPECTRAL_FLATNESS_BIN_MASK: Boolean (0 or 1).
        Whether or not to preprocess the `spectral_flatness_bin_mask`. 
        If `True`,  segment longer than 10 frames that contains pitch (based on spectral flatness), but having 0 mean, 
        removed from the pitch signal (set to 0 in the binary mask).

    Returns
    -------
    spectral_flatness_bin_mask_block: 1D binary array.
        Mask of size n_frames_fft. 1 if the frame has lower spectral flatness than the threshold (e.g. 05), 0 otherwise. 
        
    """
    spectral_flatness_bin_mask_ = deepcopy(spectral_flatness_bin_mask)

    if n_frames_fft == len(spectral_flatness_bin_mask_)+1:
        np.append(spectral_flatness_bin_mask_, spectral_flatness_bin_mask_[n_frames_fft-1])  
        
    # Post-process the binary pitch signal. 
    if POST_PROCESS_SPECTRAL_FLATNESS_BIN_MASK == 0:
        sign_pv=0
        for i in range(0, n_frames_fft):

            if ( spectral_flatness_bin_mask_[i]==1) and (sign_pv==0):
                nstart, sign_pv =i, 1
            elif ( (spectral_flatness_bin_mask_[i] == 0) or (i==n_frames_fft-1) ) and (sign_pv==1):
                nstop=i
                if i==n_frames_fft-1:
                    nstop=i+1
                sign_pv=0
                pitchseg=np.zeros(nstop-nstart)
                for j in range (nstart, nstop):   
                    pitchseg[j-nstart]=spectral_flatness[j];
                    
                # If there is a segment longer than 10 frames that contains pitch (based on spectral flatness), but having 0 mean
                # are from the pitch signal (set to 0 in the binary mask).
                if (sum(np.abs( np.round( pitchseg-np.average(pitchseg) ) ))==0)  and (nstop-nstart+1>=10):
                    spectral_flatness_bin_mask_[range(nstart,nstop)]=0 
                    
    # Pitch segments are extended by 60 frames at the beginning and end.
    sign_pv=0
    spectral_flatness_bin_mask_block=deepcopy(spectral_flatness_bin_mask_)   
    for i in range(0, n_frames_fft):

        # Start of a segment of pitch
        if (spectral_flatness_bin_mask_[i]==1) and (sign_pv==0):
            nstart, sign_pv= i, 1
            spectral_flatness_bin_mask_block[range(max([nstart-60,0]), nstart+1)]=1
            
        # End of a segment of pitch
        elif ( (spectral_flatness_bin_mask_[i] ==0) or (i==n_frames_fft-1 )) and (sign_pv==1):
            nstop, sign_pv= i, 0
            spectral_flatness_bin_mask_block[range(nstop, np.amin([nstop+60,n_frames_fft-1])+1 )]=1 
            
    return spectral_flatness_bin_mask_block 

def apply_high_pass_filter(signal):
    """Filter the signal by a high-pass ﬁlter to remove the DC component and low frequency noise. 
        Cut-off frequencyis set to 60 Hz.

    Parameters
    ----------
    signal : 1D array.
        Array of the audio file.

    Returns
    -------
    filtered_signal : 1D array.
        Filtered array of the audio file.
        
    """
    
    # A ﬁrst-order high-pass ﬁlter with a cutoﬀ frequency of 60 Hz is applied to remove low-frequency noise.
    b=np.array([0.9770,   -0.9770])
    a=np.array([1.0000,   -0.9540])

    filtered_signal=lfilter(b, a, signal, axis=0)

    return filtered_signal

def snre_highenergy(filtered_signal, n_frames_fft, samples_per_frame, samples_between_frames, spectral_flatness_bin_mask, ENERGYFLOOR, NESEG, D_SMOOTHING_LEFT, D_SMOOTHING_RIGHT, SEGMENT_THRESHOLD):
    """ Perform a first pass denoising step. High-energy segments are 
        detected by using an a posteriori SNR weighted energy diﬀerence measure.
        Frames containing noise are outputed so that their samples are subsequently set to zero.
        The energyu of the frames are also outputed for `snre_vad` function, as it is computationally 
        more expensive that the other operations.
    
    Parameters
    ----------
    filtered_signal : 1D binary array.
        Filtered audio signal (with a first-order high-pass filter).
    n_frames_fft: int.
        Number of frames of the fft signal (floor).
    samples_per_frame: int.
        Number of samples per frames.
    samples_between_frames: int
        Number of samples between frames.
    spectral_flatness_bin_mask : 1D binary array.
        Mask of size n_frames_fft. 1 if the frame has lower spectral flatness than the threshold (e.g. 05), 0 otherwise.   
    ENERGYFLOOR: float.
        Minimum possible energy.
    NESEG: int. Default=200.
        Size in frame of a super-segment.
        TODO

    Returns
    -------
    energy: 1D array dtype: float.
        Energy of each of the frames. 
    noise_samples_coordinates: 2D array dtype: int
        Array of start and stop indexes of the segments which don't contain pitch.
    noise_seg: 1D binary array.
        Binary array containing 1 if a "segment" is labeled as noisy, 0 otherwise.
    numer_of_noisy_segments: int.
        Number of segments containing noise.
    """
    
    ## ---*******- important *******
    #here [0] index array element has  not used  PROBABLY BECAUSE OF MATLAB INDEX !

    filtered_signal_ = deepcopy(filtered_signal)
    spectral_flatness_bin_mask_ = deepcopy(spectral_flatness_bin_mask)

    filtered_signal_ = np.insert(filtered_signal_,0,'inf')
    spectral_flatness_bin_mask_ = np.insert(spectral_flatness_bin_mask_,0,'inf')

    # Compute the energy of each frames of the audio signal.
    energy=np.zeros(n_frames_fft,  dtype='float64');energy=np.insert(energy,0,'inf')

    # For all the frames
    for i in range(1, n_frames_fft+1):

        # for all the samples in a frame
        for j in range(1, samples_per_frame+1):
             energy[i]=energy[i]+np.square(filtered_signal_[(i-1)*samples_between_frames+j])

        # Every frame having less than ENERGYFLOOR energy are set to that threshold.
        if np.less_equal(energy[i], ENERGYFLOOR):
             energy[i]=ENERGYFLOOR
                
                
    # Init. the e_v(m)_tilde i.e the estimated noise energy
    emin=np.ones(n_frames_fft);emin=np.insert(emin,0,'inf')

    # Check that the nuber of frames is at least the size of a super-segment
    if np.less(n_frames_fft, NESEG):
        NESEG=n_frames_fft

    # We compute the noise energy for all the super-segment
    for i in range(1, int(np.floor(n_frames_fft/NESEG))+1):

        # eY is the sorted energy of each frame of that super-segment
        eY=np.sort(energy[range((i-1)*NESEG+1, (i*NESEG)+1)]);eY=np.insert(eY,0,'inf')

        #For each super-segment x(p), the noise energy ev(p) is calculated as the energy of the frame ranked at 10% of lowest energy within the super-segment. TODO: see if we change ey by ev ...
        # In other words, the energy of each frame of each super segment is set at the 10% percentile of the energy of each frame of that super-segment
        emin[range( (i-1)*NESEG+1, i*NESEG+1)]=eY[int(np.floor(NESEG*0.1))]

        # Except for the initial super-segment, we compute the noise energy e_v(p)_tilde as a smoothed version 
        # of e_v(p) with a forgetting factor of .9.
        if np.not_equal(i, 1):
             emin[range((i-1)*NESEG+1,i*NESEG+1)]=0.9*emin[(i-1)*NESEG]+0.1*emin[(i-1)*NESEG+1]

    # We also need to do it for the part that does not fit into a partition of super-segment
    if np.not_equal(i*NESEG, n_frames_fft):
        eY=np.sort(energy[range((i-1)*NESEG+1, n_frames_fft+1)])
        eY=np.insert(eY,0,'inf')

        emin[range(i*NESEG+1,n_frames_fft+1)]=eY[int(np.floor((n_frames_fft-(i-1)*NESEG)*0.1))]
        emin[range(i*NESEG+1,n_frames_fft+1)]=0.9*emin[i*NESEG]+0.1*emin[i*NESEG+1]

    #### Compute the a posteriori SNR weighted energy difference
    D=np.zeros(n_frames_fft);D=np.insert(D,0,'inf')

    postsnr=np.zeros(n_frames_fft);postsnr=np.insert(postsnr,0,'inf')

    # Compute the a posteriori SNR weighted energy difference for each of the frames.
    for i in range(2, n_frames_fft+1):
        postsnr[i] = np.log10(energy[i])-np.log10(emin[i])
        if np.less(postsnr[i],0):
             postsnr[i]=0

        D[i]=np.sqrt(np.abs(energy[i]-energy[i-1])*postsnr[i])
    D[1]=D[2]

    ################## Central-Smooth the a posteriori SNR weighted energy difference

    # Augment the size of D (by setting D_SMOOTHING_LEFT times the first value at the beginning of the signal), to do the central-smoothing
    tm1 = np.hstack((np.ones(D_SMOOTHING_LEFT)*D[1], D[1:len(D)]))

    # Augment the size of D (by setting D_SMOOTHING_LEFT times the last value at the end of the signal), to do the central-smoothing
    Dexp = np.hstack((tm1, np.ones(D_SMOOTHING_RIGHT)*D[n_frames_fft] ));Dexp = np.insert(Dexp,0,'inf')


    Dsmth=np.zeros(n_frames_fft, dtype='float64');Dsmth=np.insert(Dsmth,0,'inf')

    Dsmth_max=deepcopy(Dsmth)

    for i in range(1,n_frames_fft+1):
        Dsmth[i]=sum(Dexp[range(i, i+D_SMOOTHING_LEFT+D_SMOOTHING_RIGHT+1)])


    # For all the super-segments, we compute the maximum of the a posteriori SNR weighted energy difference, and set it to all the frames
    for i in range(1, int(np.floor(n_frames_fft/NESEG))+1):
        Dsmth_max[range((i-1)*NESEG+1, i*NESEG+1)]= np.amax(energy[range((i-1)*NESEG+1, i*NESEG+1)]);  #np.amax(Dsmth[range((i-1)*NESEG+1, i*NESEG+1)]) TODO check why not using this ? it lower the scale

    # We also need to do it for the part that does not fit into a partition of super-segment
    if np.not_equal(i*NESEG, n_frames_fft):
        Dsmth_max[range(i*NESEG+1, n_frames_fft+1)]=np.amax(energy[range((i-1)*NESEG+1, n_frames_fft+1)])     #np.amax(Dsmth[range((i-1)*NESEG+1, n_frames_fft+1)])


        
    # For all the frames we classify the high-energy sframes
    snre_vad = np.greater(Dsmth, Dsmth_max*SEGMENT_THRESHOLD).astype(float)
    snre_vad = np.insert(snre_vad,0,'inf')

    
    #################### Block-based processing: we use the binary mask of high-energy frames
    # Consecutive high-energy frames are grouped together to form high-energy segments.  
    # (e) Within a high-energy segment, if no more than two pitch frames are found, the segment is classiﬁed as noise, 
    # and the samples of the segment are set to zero.

    # Mask that indicate a segment of noise
    noise_seg=np.zeros(int(np.floor(n_frames_fft/1.6))) # TODO: WHat does this 1.6 ??? 
    noise_seg=np.insert(noise_seg,0,'inf')

    # Start and end timestamps of noisy segments in the referentiel of the frames.
    noise_samples_coordinates=np.zeros((n_frames_fft,2))
    n_noise_samples_coordinates=-1
    sign_vad = 0


    # For each of the frames, we look for consecutive frames of high-energy
    for i in range(1, n_frames_fft+1):

        # start of a segment of high-energy
        if (snre_vad[i] == 1) and (sign_vad == 0): 
            sign_vad = 1
            nstart=i

        # end of a segment
        elif ((snre_vad[i] ==0) or (i==n_frames_fft)) and (sign_vad == 1): 
            sign_vad = 0
            nstop=i-1
            if np.equal(sum(spectral_flatness_bin_mask_[range(nstart, nstop+1)]), 0):
                noise_seg[range(int(np.round(nstart/1.6)), int(np.floor(nstop/1.6))+1)] = 1
                n_noise_samples_coordinates=n_noise_samples_coordinates+1
                noise_samples_coordinates[n_noise_samples_coordinates,:]=np.array([(nstart-1)*samples_between_frames+1, nstop*samples_between_frames])

    noise_samples_coordinates=noise_samples_coordinates[:n_noise_samples_coordinates+1,]

    #syn  from [0] index
    noise_samples_coordinates=noise_samples_coordinates-1
    noise_seg=noise_seg[1:len(noise_seg)]
        
    return energy, noise_samples_coordinates, noise_seg, len(noise_samples_coordinates)   

def snre_vad(filtered_signal, n_frames_fft, samples_per_frame, samples_between_frames, spectral_flatness_bin_mask, spectral_flatness_bin_mask_block, energy, ENERGYFLOOR, VAD_THRESHOLD, D_SMOOTHING_LEFT, D_SMOOTHING_RIGHT):
    """ Perform the voice Activity Detection based on the computed spectral flatness and the filtered signal energy.
        See section 3.4 of the paper for more details. 
    
    Parameters
    ----------
    filtered_signal : 1D binary array.
        Filtered audio signal (with a first-order high-pass filter).
    n_frames_fft: int.
        Number of frames of the fft signal (floor).
    samples_per_frame: int.
        Number of samples per frames.
    samples_between_frames: int
        Number of samples between frames.
    spectral_flatness_bin_mask : 1D binary array.
        Mask of size n_frames_fft. 1 if the frame has lower spectral flatness than the threshold (e.g. 05), 0 otherwise.    
    spectral_flatness_bin_mask_block: 1D binary array.
        Same as spectral_flatness_bin_mask but extended to 60 frames before and after each segments (and an optionnal post-processing step of the
        binary mask).
    energy: 1D array dtype: float.
        Energy of each of the frames. 
    ENERGYFLOOR: float.
        Minimum possible energy.
    VAD_THRESHOLD: float.
        Voice Activity Detection Threshold. Set to .4 in the paper, and refered as beta.

    Returns
    -------
    vad_seg: 1D binary array.
        Mask of size n_frames_fft. 1 if the frame constains speech, 0 otherwise.  
    """
    
    
    ## ---*******- important *******
    #here [0] index array element is not used


    # Copy and init.
    filtered_signal_ = deepcopy(filtered_signal); filtered_signal_ = np.insert(filtered_signal_,0,'inf')
    spectral_flatness_bin_mask_ = deepcopy(spectral_flatness_bin_mask); spectral_flatness_bin_mask_ = np.insert(spectral_flatness_bin_mask_,0,'inf')
    spectral_flatness_bin_mask_block_ = deepcopy(spectral_flatness_bin_mask_block); spectral_flatness_bin_mask_block_ = np.insert(spectral_flatness_bin_mask_block_,0,'inf')

    # TODO: name oall of those. 
    Dsmth=np.zeros(n_frames_fft, dtype='float64'); Dsmth=np.insert(Dsmth,0,'inf')    
    D=np.zeros(n_frames_fft); D=np.insert(D,0,'inf')
    segsnr=np.zeros(n_frames_fft); segsnr=np.insert(segsnr,0,'inf')
    postsnr=np.zeros(n_frames_fft, dtype='float64'); postsnr=np.insert(postsnr,0,'inf')
    snre_vad=np.zeros(n_frames_fft); snre_vad=np.insert(snre_vad,0,'inf')

    sign_pv=0
    segsnrsmth=1
    sign_segsnr=0

    for i in range(1, n_frames_fft+1):

        if (spectral_flatness_bin_mask_block_[i]==1) and (sign_pv==0):
            nstart=i
            sign_pv=1

        elif ( (spectral_flatness_bin_mask_block_[i]==0) or (i==n_frames_fft) ) and (sign_pv==1): 

            nstop=i-1
            if i==n_frames_fft:
                nstop=i
            sign_pv=0

            datai=filtered_signal_[range( (nstart-1)*samples_between_frames+1, (nstop-1)*samples_between_frames+samples_per_frame-samples_between_frames+1) ]
            datai=np.insert(datai,0,'inf')

            # Calculate a posteriori SNR weighted energy diﬀerence for each frame of the extended pitch segments (based on spectral flatness)
            for j in range(nstart, nstop-1+1): 

                for h in range(1, samples_per_frame+1):
                    energy[j]=energy[j]+ np.square(datai[(j-nstart)*samples_between_frames+h] )

                if np.less_equal(energy[j], ENERGYFLOOR):
                    energy[j]=ENERGYFLOOR
            energy[nstop]=energy[nstop-1]

            eY=np.sort(energy[range(nstart, nstop+1)] );eY=np.insert(eY,0,'inf') #as [0] is discarded
            emin=eY[int(np.floor((nstop-nstart+1)*0.1))]

            for j in range(nstart+1, nstop+1):

                postsnr[j] = math.log10(energy[j]) - math.log10(emin)

                if np.less(postsnr[j], 0):
                    postsnr[j]=0

                D[j]=math.sqrt(np.abs(energy[j]-energy[j-1])*postsnr[j] )

            D[nstart]=D[nstart+1]

            ################## Central-Smooth the a posteriori SNR weighted energy difference

            # Augment the size of D (by setting D_SMOOTHING_LEFT times the first value at the beginning of the signal), to do the central-smoothing
            tm1 = np.hstack((np.ones(D_SMOOTHING_LEFT)*D[nstart], D[range(nstart, nstop+1)]))

            # Augment the size of D (by setting D_SMOOTHING_LEFT times the last value at the end of the signal), to do the central-smoothing
            Dexp = np.hstack((tm1, np.ones(D_SMOOTHING_RIGHT)*D[nstop] ));Dexp = np.insert(Dexp,0,'inf')

            for j in range(0, nstop-nstart+1):
                Dsmth[nstart+j]=sum(Dexp[range(j+1, j+D_SMOOTHING_LEFT+D_SMOOTHING_RIGHT+1)])

            Dsmth_thres=sum(Dsmth[range(nstart, nstop+1)]*spectral_flatness_bin_mask_[range(nstart, nstop+1)])/sum(spectral_flatness_bin_mask_[range(nstart,nstop+1)])

            # Classify the segments that contains speech based on the `VAD_THRESHOLD` and the a posteriory weighted difference
            snre_vad[range(nstart, nstop+1)] = np.greater(Dsmth[range(nstart, nstop+1)], Dsmth_thres*VAD_THRESHOLD).astype(float)

    # Deep copy    
    pv_vad=deepcopy(snre_vad)       

    # First, frames that are 33 frames away from the pitch segment to the left and 47 frames
    # away to the right are classiﬁed as non-speech, regardless the VAD results above, 
    # which covers 95% of the cases based on the a few speech ﬁles. cf paper section 3.4.d
    nexpl=33
    nexpr=47 # % 29 and 39, estimated statistically, 95% ; 33, 47 %98 from voicebox pitch

    sign_vad=0
    for i in range(1, n_frames_fft+1):
        if (snre_vad[i]==1) and (sign_vad==0):
            nstart=i
            sign_vad=1
        elif ((snre_vad[i]==0) or (i==n_frames_fft)) and (sign_vad==1):
            nstop=i-1
            if i==n_frames_fft:
                nstop=i
            sign_vad=0
            for j in range(nstart, nstop+1):
                if spectral_flatness_bin_mask_[j]==1:
                    break

            pv_vad[range(nstart, np.max([j-nexpl-1,1])+1)]=0

            for j in range(0, nstop-nstart+1):
                if spectral_flatness_bin_mask_[nstop-j]==1:
                    break

            pv_vad[range(nstop-j+1+nexpr,nstop+1)]=0

    # Frames that are within 5 frames to the left and 12 frames to the right of the pitch segments
    # are classiﬁed as speech, again regardless the VAD results, which laverage_energy_speech_segmentss out 5% of the cases 
    # based on the a few speech ﬁles.
    nexpl =5; nexpr=12 #; % 9 and 13, estimated statistically 5%; 5, 12 %2 for voicebox pitch

    sign_vad=0
    for i in range(1,n_frames_fft+1):
        if (snre_vad[i]==1) and (sign_vad==0):
            nstart=i
            sign_vad=1
        elif ((snre_vad[i]==0) or (i==n_frames_fft) ) and (sign_vad==1):
            nstop=i-1  
            if i==n_frames_fft:
                nstop=i
            sign_vad=0

            if np.greater(sum(spectral_flatness_bin_mask_[range(nstart,nstop+1)]), 4):
                for j in range(nstart,nstop+1):
                    if spectral_flatness_bin_mask_[j]==1:
                        break

                pv_vad[range(np.maximum(j-nexpl,1),j-1+1)]=1
                for j in range(0,nstop-nstart+1):
                    if spectral_flatness_bin_mask_[nstop-j]==1:
                        break
                pv_vad[range(nstop-j+1,min(nstop-j+nexpr,n_frames_fft)+1)]=1

            # Segments with an average energy below 0.001 are removed.
            esegment=sum(energy[range(nstart,nstop+1)])/(nstop-nstart+1)
            if np.less(esegment, 0.001):
                pv_vad[range(nstart, nstop+1)]=0

            # Segments with an average energy below 0.001 are removed.
            if np.less_equal(sum(spectral_flatness_bin_mask_[range(nstart,nstop+1)]),  2):
                pv_vad[range(nstart,nstop+1)] = 0


    # Compute cumulative sum of all speech frames
    sign_vad=0
    cumulative_energy_speech_segments=0
    for i in range(1,n_frames_fft+1):
        if (pv_vad[i]==1) and (sign_vad==0):
            nstart=i
            sign_vad=1
        elif ((pv_vad[i]==0) or (i==n_frames_fft)) and (sign_vad==1):
            nstop=i-1
            if i==n_frames_fft:
                nstop=i
            sign_vad=0
            cumulative_energy_speech_segments=cumulative_energy_speech_segments+sum(energy[range(nstart, nstop+1)])

    # Copute the average energy of the speech segments
    eps = np.finfo(float).eps
    average_energy_speech_segments=cumulative_energy_speech_segments/(sum(pv_vad[1:len(pv_vad)])+eps) # except [0] index 'inf'


    # Segments with energy below 0.05 times the overall energy are removed
    sign_vad=0
    for i in range(1,n_frames_fft+1):
        if (pv_vad[i]==1) and (sign_vad==0):
            nstart=i
            sign_vad=1
        elif ((pv_vad[i]==0) or (i==n_frames_fft)) and (sign_vad==1):
            nstop=i-1
            if i==n_frames_fft:
                nstop=i
            sign_vad=0

            # Segments with energy below 0.05 times the overall energy is removed.
            #if np.less(sum(energy[range(nstart,nstop+1)])/(nstop-nstart+1), average_energy_speech_segments*0.05):
            #    pv_vad[range(nstart,nstop+1)] = 0

    # Compute start and stop frames of every speech segments
    sign_vad=0
    vad_seg=np.zeros((n_frames_fft,2), dtype="int64")
    n_vad_seg=-1 #for indexing array
    for i in range(1,n_frames_fft+1):
        if (pv_vad[i]==1) and (sign_vad==0):
            nstart=i
            sign_vad=1
        elif ((pv_vad[i]==0) or (i==n_frames_fft)) and (sign_vad==1):
            nstop=i-1
            sign_vad=0
            n_vad_seg=n_vad_seg+1
            vad_seg[n_vad_seg,:]=np.array([nstart, nstop])

    vad_seg=vad_seg[:n_vad_seg+1,]
    vad_seg = vad_seg - 1

    # Compute final binary mask of speech signal 
    final_vad_seg=np.zeros(n_frames_fft, dtype="int64")
    for i in range(len(vad_seg)):  
        final_vad_seg[range(vad_seg[i,0], vad_seg[i,1]+1)]=1
    
    return final_vad_seg, vad_seg


def idx2time(n, frame_per_second):
    #idx should be in the reference of the entire video, ie between 0 and self.dataset.n_frames
    if int(n/frame_per_second//60)==0:
        return('{:.2f}s'.format(n/frame_per_second%60))
    else:
        return('{} | {}m {:.2f}s'.format(int(n/frame_per_second//60), n/frame_per_second%60))


def create_render_timeline(signal, t_max):

    n = signal.shape[0]; frame_per_second = int(n/t_max)
    xticklabels = [idx2time(i, frame_per_second) for i in range(0,n)[0:None:n//20]]
    xticklocation = np.linspace(0, n, n)[0:None:n//20]
    return xticklabels, xticklocation
    
def plot_vad(signal, final_vad_seg, vad_seg_timestamps, t_max, path='./data/output_0.png'):
    
    #fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(20, 12))
    
    fig = plt.figure(figsize=(10, 3))
    ax1=plt.gca()


    ############ ploting the Waveform ############

    # Create x-y ticks
    n = signal.shape[0]; frame_per_second = int(n/t_max)
    xticklabels = [idx2time(i, frame_per_second) for i in range(0,n)[0:None:n//20]]
    xticklocation = np.linspace(0, n, n)[0:None:n//20]

    ax1.plot(signal, color=colors[0]); ax1.set_xlim(0, t_max)
    ax1.set_xticks(xticklocation);ax1.set_xticklabels(xticklabels, rotation = 90)
    ax1.set_yticks([]);ax1.set_yticklabels([], rotation = 90)
    ax1.set_title("Waveform of the input + VAD", weight='bold', fontsize=18)
    ax1.set_xlabel("Time [s]", fontsize=13)
    ymin = ax1.get_ylim()[0]; ymax = ax1.get_ylim()[1]
    ax1.fill_between(vad_seg_timestamps[0], y1=ymin, y2=ymax, color=colors[1], alpha=.5, label="Speech segment")
    for start_stop_tuple in vad_seg_timestamps[1:]:
        ax1.fill_between(start_stop_tuple, y1=ymin, y2=ymax, color=colors[1], alpha=.5)
    ax1.legend(prop={'size':10})

    if False:
        ############ ploting the heatmap ############

        # Create x-y ticks
        n = final_vad_seg.shape[0]; frame_per_second = np.ceil(n/t_max)
        xticklabels = [idx2time(i, frame_per_second) for i in range(0,n)][0:None:n//20]
        xticklocation = np.linspace(0, n, n)[0:None:n//20]

        ax2 = sns.heatmap(final_vad_seg[None,:], yticklabels=False, cmap=cmap, ax=ax2, cbar_kws={'pad':.2, 'orientation' : 'horizontal'})
        ax2.set_title("Voice Activity Detection output", weight='bold', fontsize=18)
        plt.xlabel("Time [s]", weight='bold', fontsize=13)
        ax2.set_xticks(xticklocation);ax2.set_xticklabels(xticklabels, rotation = 90)
        colorbar = ax2.collections[0].colorbar;r = np.unique(final_vad_seg);n = len(r)
        colorbar.set_ticks([r.min() + (r.max()-r.min())/n*(0.5 + i) for i in range(n)])
        colorbar.set_ticklabels([v for i, v  in {0:"No Human Voice", 1:"Human Voice"}.items() if i in r], weight='bold', fontsize=13)   

    # Organize layout
    plt.tight_layout()

    # save the results for archival
    plt.savefig(path)
    
    return 